/*
 * Main.c
 *
 *  Created on: 04.01.2014
 *      Author: vgarske
 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>

// #define IS_AMMETER

static uint8_t Digits[4] = { 0, 0, 0, 0 };

#define SEGA	(1 << PB6)
#define SEGB	(1 << PB0)
#define SEGC	(1 << PB2)
#define SEGD	(1 << PB5)
#define SEGE	(1 << PB4)
#define SEGF	(1 << PB7)
#define SEGG	(1 << PB1)
#define SEGDP	(1 << PB3)

static uint8_t ValueTo7Seg[10] =
{
	(uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF)),
	(uint8_t) (~(SEGB | SEGC)),
	(uint8_t) (~(SEGA | SEGB | SEGD | SEGE | SEGG)),
	(uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGG)),
	(uint8_t) (~(SEGB | SEGC | SEGF | SEGG)),
	(uint8_t) (~(SEGA | SEGC | SEGD | SEGF | SEGG)),
	(uint8_t) (~(SEGA | SEGC | SEGD | SEGE | SEGF | SEGG)),
	(uint8_t) (~(SEGA | SEGB | SEGC)),
	(uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF | SEGG)),
	(uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGF | SEGG)),
};

#define DECIMAL_POINT	~SEGDP

#define BLANK			0xff
#define MINUS			((uint8_t) ~SEGG)

static uint8_t ButtonPressed = 0;
static uint16_t PressedCounter;

typedef enum
{
	NO_MESSAGE,
	BTN_PRESSED_SHORT,
	BTN_PRESSED_LONG
} MESSAGE;
static MESSAGE WaitingMessage = NO_MESSAGE;

static uint16_t	ConvertMinimumADC, ConvertZeroADC, ConvertMaximumADC;

// ============================================================================
//  ISR stubs
// ============================================================================

EMPTY_INTERRUPT (TIMER2_COMPA_vect)
EMPTY_INTERRUPT (TIMER2_OVF_vect)
EMPTY_INTERRUPT (ADC_vect)

// ============================================================================
//  EEProm
// ============================================================================

#define EEPROM_SIGNATURE		0x566f4761
#define EEPROM_ADDR_MINIMUM		0
#define EEPROM_ADDR_ZERO		2
#define EEPROM_ADDR_MAXIMUM		4
#define EEPROM_ADDR_SIGNATURE	6

static void StoreEEProm (void)
{
	eeprom_write_word ((uint16_t*) EEPROM_ADDR_MINIMUM, ConvertMinimumADC);
	eeprom_write_word ((uint16_t*) EEPROM_ADDR_ZERO, ConvertZeroADC);
	eeprom_write_word ((uint16_t*) EEPROM_ADDR_MAXIMUM, ConvertMaximumADC);
	eeprom_write_dword ((uint32_t*) EEPROM_ADDR_SIGNATURE, EEPROM_SIGNATURE);
}

static void RetrieveEEProm (void)
{
	uint32_t signature;

	ConvertMinimumADC = eeprom_read_word ((uint16_t*) EEPROM_ADDR_MINIMUM);
	ConvertZeroADC = eeprom_read_word ((uint16_t*) EEPROM_ADDR_ZERO);
	ConvertMaximumADC = eeprom_read_word ((uint16_t*) EEPROM_ADDR_MAXIMUM);
	signature = eeprom_read_dword ((uint32_t*) EEPROM_ADDR_SIGNATURE);
	if (signature != EEPROM_SIGNATURE)
	{
#ifdef IS_AMMETER
		ConvertMinimumADC = (uint16_t) ((uint32_t) 4096 * 2450 / 250);
		ConvertZeroADC = (uint16_t) ((uint32_t) 4096 * ((2450 - 50) / 2) / 250);
		ConvertMaximumADC = (uint16_t) ((uint32_t) 4096 * 50 / 250);
#else
		ConvertMinimumADC = (uint16_t) ((uint32_t) 4096 * 2450 / 3000);
		ConvertZeroADC = (uint16_t) ((uint32_t) 4096 * ((2450 - 50) / 2) / 3000);
		ConvertMaximumADC = (uint16_t) ((uint32_t) 4096 * 50 / 3000);
#endif
		StoreEEProm ();
	}
}

// ============================================================================
//  Conversion functions
// ============================================================================

static uint16_t CalculateBattVoltage (uint16_t adcValue)
{
//	return ((uint16_t) (100.0 * (adcValue * 2.5f / 4096) / (10000.0 / 16800)));
	return ((uint16_t) ((uint32_t) 420 * adcValue / 4096));
}

static void TranslateMeasurement (uint16_t adcValue)
{
	int16_t result, divider;

	if (adcValue < 4)
	{
		Digits[0] = BLANK;
		Digits[1] = (uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF));
		Digits[2] = (uint8_t) (~(SEGD | SEGE | SEGF));
		Digits[3] = BLANK;
		return;
	}

	if (adcValue >= 0xffc)
	{
		Digits[0] = MINUS;
		Digits[1] = (uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF));
		Digits[2] = (uint8_t) (~(SEGD | SEGE | SEGF));
		Digits[3] = BLANK;
		return;
	}

	if (adcValue > ConvertZeroADC)
	{
#ifdef IS_AMMETER
		result = (int32_t) -250 * (adcValue - ConvertZeroADC) / (ConvertMinimumADC - ConvertZeroADC);
#else
		result = (int32_t) -3000 * (adcValue - ConvertZeroADC) / (ConvertMinimumADC - ConvertZeroADC);
#endif
	} else
	{
#ifdef IS_AMMETER
		result = (uint32_t) 250 * (ConvertZeroADC - adcValue) / (ConvertZeroADC - ConvertMaximumADC);
#else
		result = (uint32_t) 3000 * (ConvertZeroADC - adcValue) / (ConvertZeroADC - ConvertMaximumADC);
#endif
	}

	if (result < 0)
	{
		Digits[0] = MINUS;
		result = -result;
	} else
	{
		Digits[0] = BLANK;
	}

	divider = 1000;
	if (result < 1000)
	{
		divider = 100;
	}

	for (uint8_t digit = 1; digit < 4; digit++)
	{
		Digits[digit] = ValueTo7Seg[result / divider];
		if (divider == 100)
		{
			Digits[digit] &= DECIMAL_POINT;
		}
		result %= divider;
		divider /= 10;
	}
}

static void TranslateMeasurementRaw (uint16_t adcValue)
{
	int16_t result, divider;

	result = (uint32_t) 2500 * adcValue / 4096;
	divider = 1000;

	for (uint8_t digit = 0; digit < 4; digit++)
	{
		Digits[digit] = ValueTo7Seg[result / divider];
		result %= divider;
		divider /= 10;
	}
	Digits[0] &= DECIMAL_POINT;
}

// ============================================================================
//  ADC
// ============================================================================

static void StartADC (void)
{
	ADMUX = (0 << REFS1) | (0 << REFS0) | (0 << ADLAR);
	ADCSRA = (1 << ADEN) | (1 << ADIE) | (0 << ADPS2) | (0 << ADPS1) | (0 << ADPS0);
}

static void StopADC (void)
{
	ADCSRA = (0 << ADEN);
}

static uint16_t MeasureADC (int channel)
{
	uint16_t accumulate = 0;

	ADMUX = (ADMUX & 0xf0) | channel;
	set_sleep_mode (SLEEP_MODE_ADC);
	sleep_enable ();

	for (uint8_t u = 0; u < 16; u++)
	{
		sleep_cpu ();
		accumulate += ADC;
	}

	sleep_disable ();

	return (accumulate / 4);
}

// ============================================================================
//  Timer
// ============================================================================

typedef enum
{
	TIMER_2500USEC,
	TIMER_262MSEC
} DURATION;

static void StartTimer (DURATION duration)
{
	TCNT2 = 0;
	switch (duration)
	{
		case TIMER_2500USEC:
			OCR2A = F_CPU / (4 * 100) / 32; // 4 digits, 100 Hz refresh rate, 1:32 prescaler
			TIMSK2 = (1 << OCIE2A);
			TCCR2A = (1 << WGM21) | (0 << WGM20);
			TCCR2B = (0 << WGM22) | (0 << CS22) | (1 << CS21) | (1 << CS20);
			break;

		case TIMER_262MSEC:
			TIMSK2 = (1 << TOIE2); // 256 clks, 1:1024 prescaler
			TCCR2A = (0 << WGM21) | (0 << WGM20);
			TCCR2B = (0 << WGM22) | (1 << CS22) | (1 << CS21) | (1 << CS20);
			break;
	}
}

static void StopTimer (void)
{
	TCCR2B = (0 << CS22) | (0 << CS21) | (0 << CS20);
}

static void WaitForTimer (void)
{
	set_sleep_mode (SLEEP_MODE_PWR_SAVE);
	sleep_mode ();
}

// ============================================================================
//  Button
// ============================================================================

static void SampleButton (void)
{
	uint8_t uPrev = ButtonPressed;

	ButtonPressed = (ButtonPressed << 1) | ((PIND & (1 << PIND1)) ? 0 : 1);

	if (ButtonPressed == 0xff)
	{
		if (uPrev == 0xff)
		{
			PressedCounter++;
		} else
		{
			PressedCounter = 0;
		}
	}

	if (ButtonPressed == 0x00 && uPrev != 0x00)
	{
		if (PressedCounter < 4 * 50 * 4) // 4 seconds
		{
			WaitingMessage = BTN_PRESSED_SHORT;
		} else
		{
			WaitingMessage = BTN_PRESSED_LONG;
		}
	}
}

static MESSAGE GetButtonMessage (void)
{
	MESSAGE retrievedMessage;

	retrievedMessage = WaitingMessage;
	WaitingMessage = NO_MESSAGE;

	return (retrievedMessage);
}

// ============================================================================
//  Display
// ============================================================================

static void ShowDigit (uint8_t digit)
{
	uint8_t colSelector;

	PORTB = 0xff;
	PORTD |= (1 << PD2) | (1 << PD3) | (1 << PD4) | (1 << PD5);

	switch (digit)
	{
		case 0: 	colSelector = ~(1 << PD4); break;
		case 1: 	colSelector = ~(1 << PD3); break;
		case 2: 	colSelector = ~(1 << PD2); break;
		case 3: 	colSelector = ~(1 << PD5); break;
		default:	return;
	}

	PORTB = Digits[digit];
	PORTD &= colSelector;
}

static void ShowFourDigits (void)
{
	for (uint8_t u = 0; u < 4; u++)
	{
		ShowDigit (u);
		StartTimer (TIMER_2500USEC);
		SampleButton ();
		WaitForTimer ();
	}
	ShowDigit (4);
	StopTimer ();
}

typedef enum
{
	WAIT_ONLY,
	WAIT_AND_ALLOW_BTN
} ALLOWBTN;

static uint8_t ShowText (uint8_t digit0,
	uint8_t digit1,
	uint8_t digit2,
	uint8_t digit3,
	uint16_t delay,
	ALLOWBTN allowBtn)
{
	Digits[0] = digit0;
	Digits[1] = digit1;
	Digits[2] = digit2;
	Digits[3] = digit3;
	for (uint16_t u = 0; u < delay; u++)
	{
		ShowFourDigits ();
		if (allowBtn == WAIT_AND_ALLOW_BTN && GetButtonMessage () == BTN_PRESSED_SHORT)
		{
			return (1);
		}
	}

	return (0);
}

static void ShowBatt (void)
{
	ShowText ((uint8_t) (~(SEGC | SEGD | SEGE | SEGF | SEGG)),
			  (uint8_t) (~(SEGA | SEGB | SEGC | SEGE | SEGF | SEGG)),
			  (uint8_t) (~(SEGD | SEGE | SEGF | SEGG)),
			  (uint8_t) (~(SEGD | SEGE | SEGF | SEGG)),
			  100,
			  WAIT_ONLY);

	StartADC ();
	uint16_t batt = CalculateBattVoltage (MeasureADC (0));
	StopADC ();

	Digits[0] = BLANK;
	Digits[1] = ValueTo7Seg[batt / 100] & DECIMAL_POINT;
	Digits[2] = ValueTo7Seg[(batt % 100) / 10];
	Digits[3] = ValueTo7Seg[batt % 10];

	for (uint8_t u = 0; u < 100; u++)
	{
		ShowFourDigits ();
	}

	StartTimer (TIMER_262MSEC);
	for (uint8_t u = 0; u < 2; u++)
	{
		WaitForTimer ();
	}
	StopTimer ();
}

// ============================================================================
//  Hardware init
// ============================================================================

static void InitHardware (void)
{
// I/O ports
	DDRB = BLANK;
	DDRD = (1 << DDD0) | (0 << DDD1) | (1 << DDD2) | (1 << DDD3) | (1 << DDD4) | (1 << DDD5);
// KeepOn=1, PullUp PwrBtn, 7seg cols off
	PORTD = (1 << PD0) | (1 << PD1) | (1 << PD2) | (1 << PD3) | (1 << PD4) | (1 << PD5);

// Turn off AVR modules
	ACSR = (1 << ACD); // no Analog Comparator
	DIDR0 = (1 << ADC5D) | (1 << ADC4D) | (1 << ADC3D) | (1 << ADC2D) | (1 << ADC1D) | (1 << ADC0D); // no digital buffers on port C;
	DIDR1 = (1 << AIN1D) | (1 << AIN0D); // no digital buffers on PD6 and PD7;
	PRR = (1 << PRTWI) | (1 << PRTIM0) | (1 << PRTIM1) | (1 << PRSPI) | (1 << PRUSART0);

// Wait until btn released
	while (!(PIND & (1 << PIND1))) ;

	sei ();
}

// ============================================================================
//  Main
// ============================================================================

static void Shutdown (void)
{
// Scroll text "OFF"
	ShowText (BLANK,
			  BLANK,
			  BLANK,
			  (uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF)),
			  20,
			  WAIT_ONLY);
	ShowText (BLANK,
			  BLANK,
			  (uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF)),
			  (uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  20,
			  WAIT_ONLY);
	ShowText (BLANK,
			  (uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF)),
			  (uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  (uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  20,
			  WAIT_ONLY);
	ShowText ((uint8_t) (~(SEGA | SEGB | SEGC | SEGD | SEGE | SEGF)),
			  (uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  (uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  BLANK,
			  20,
			  WAIT_ONLY);
	ShowText ((uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  (uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  BLANK,
			  BLANK,
			  20,
			  WAIT_ONLY);
	ShowText ((uint8_t) (~(SEGA | SEGE | SEGF | SEGG)),
			  BLANK,
			  BLANK,
			  BLANK,
			  20,
			  WAIT_ONLY);

	set_sleep_mode (SLEEP_MODE_PWR_DOWN);
	PORTD &= ~(1 << PD0);
	for (;;)
	{
		sleep_mode ();
	}
}

static void Calibrate (void)
{
	uint16_t newZero, newMax;

	if (!ShowText ((uint8_t) (~(SEGA | SEGD | SEGE | SEGF)),
			       (uint8_t) (~(SEGA | SEGB | SEGC | SEGE | SEGF | SEGG)),
			       (uint8_t) (~(SEGD | SEGE | SEGF)),
				   BLANK,
				   500,
				   WAIT_AND_ALLOW_BTN))
	{
		return;
	}

	if (!ShowText (BLANK,
				   ValueTo7Seg[0],
				   ValueTo7Seg[0] & DECIMAL_POINT,
				   ValueTo7Seg[0],
				   60 * 100, // 60 seconds
				   WAIT_AND_ALLOW_BTN))
	{
		return;
	}

	StartTimer (TIMER_262MSEC);
	for (uint8_t u = 0; u < 4; u++)
	{
		WaitForTimer ();
	}
	StopTimer ();

	StartADC ();
	newZero = MeasureADC (1);
	StopADC ();

	TranslateMeasurementRaw (newZero);
	for (uint8_t u = 0; u < 50; u++)
	{
		ShowFourDigits ();
	}

	if (!ShowText (
#ifdef IS_AMMETER
				   BLANK,
				   ValueTo7Seg[2] & DECIMAL_POINT,
				   ValueTo7Seg[5],
				   ValueTo7Seg[0],
#else
				   BLANK,
				   ValueTo7Seg[3],
				   ValueTo7Seg[0] & DECIMAL_POINT,
				   ValueTo7Seg[0],
#endif
				   60 * 100, // 60 seconds
				   WAIT_AND_ALLOW_BTN))
	{
		return;
	}

	StartTimer (TIMER_262MSEC);
	for (uint8_t u = 0; u < 4; u++)
	{
		WaitForTimer ();
	}
	StopTimer ();

	StartADC ();
	newMax = MeasureADC (1);
	StopADC ();

	TranslateMeasurementRaw (newMax);
	for (uint8_t u = 0; u < 50; u++)
	{
		ShowFourDigits ();
	}

	if (!ShowText (
#ifdef IS_AMMETER
				   MINUS,
				   ValueTo7Seg[2] & DECIMAL_POINT,
				   ValueTo7Seg[5],
				   ValueTo7Seg[0],
#else
			       MINUS,
				   ValueTo7Seg[3],
				   ValueTo7Seg[0] & DECIMAL_POINT,
				   ValueTo7Seg[0],
#endif
				   60 * 100, // 60 seconds
				   WAIT_AND_ALLOW_BTN))
	{
		return;
	}

	StartTimer (TIMER_262MSEC);
	for (uint8_t u = 0; u < 4; u++)
	{
		WaitForTimer ();
	}
	StopTimer ();

	StartADC ();
	ConvertMinimumADC = MeasureADC (1);
	StopADC ();

	TranslateMeasurementRaw (ConvertMinimumADC);
	for (uint8_t u = 0; u < 50; u++)
	{
		ShowFourDigits ();
	}

	ConvertZeroADC = newZero;
	ConvertMaximumADC = newMax;

	StoreEEProm ();
}

int main (void)
{
	InitHardware ();

	RetrieveEEProm ();

	ShowBatt ();

// Empty keyboard buffer
	GetButtonMessage ();

	for (uint32_t timer = 0; timer < (uint32_t) 10 * 60 * 100; timer++) // 10 mins @ 100 Hz
	{
// Measure
		StartADC ();
		if (CalculateBattVoltage (MeasureADC (0)) < 300)
		{
			Shutdown ();
		}

		if ((timer & 15) == 0)
		{
			uint16_t adc = MeasureADC (1);
			StopADC ();
			TranslateMeasurement (adc);
		}

// Display
		ShowFourDigits ();

		switch (GetButtonMessage ())
		{
			case BTN_PRESSED_SHORT:
				Shutdown ();
				break;

			case BTN_PRESSED_LONG:
				Calibrate ();
				timer = 0;
				break;
		}
	}

	Shutdown ();

	return (0);
}
